#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2017:
#    Gabes Jean, naparuba@gmail.com
#    Gerhard Lausser, Gerhard.Lausser@consol.de
#    Gregory Starck, g.starck@gmail.com
#    Hartmut Goebel, h.goebel@goebel-consult.de
#
# This file is part of Shinken.
#
# Shinken is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Shinken is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Shinken.  If not, see <http://www.gnu.org/licenses/>.


import base64
import cPickle
import copy
import json
import logging
import multiprocessing
import os
import signal
import socket
import sys
import threading
import time
import traceback
import urllib2
import zlib
from collections import defaultdict, deque
from multiprocessing import active_children

from shinken.basesubprocess import LookAtMyFatherThread
from shinken.brok import Brok
from shinken.daemon import IStatsInterface
from shinken.external_command import ExternalCommand
from shinken.http_client import HTTPException, HTTPExceptions, HTTPClient
from shinken.load import AvgInRange
from shinken.log import logger, get_chapter_string, LoggerFactory, get_section_string
from shinken.misc.type_hint import TYPE_CHECKING
from shinken.objects.graphitebackend import GraphiteBackend
from shinken.property import IntegerProp, PathProp
from shinken.runtime_stats.cpu_stats import cpu_stats_helper
from shinken.runtime_stats.threads_dumper import WatchDogThreadDumper
from shinken.safepickle import SafeUnpickler
from shinken.subprocess_helper.after_fork_cleanup import after_fork_cleanup
from shinken.toolbox.url_helper import BaseUrl, ShinkenBaseUrlException
from shinken.util import mem_wait_for_fork_possible, set_process_name, to_bool
from shinken.withinventorysatellite import IArbiterToInventorySatellite, WithInventorySatellite
from shinkensolutions.lib_checks.graphite import GraphiteVersionChecker
from shinkensolutions.lib_checks.graphite_metrics_counter import GRAPHITE_STATS_KEY

try:
    from statsd import StatsClient  # noqa => no need in requirements
    
    statsd = StatsClient()
except ImportError:
    StatsClient = None
    statsd = None

if TYPE_CHECKING:
    from shinken.brok import Brok
    from shinken.log import PartLogger
    from shinken.misc.type_hint import Any, List, NoReturn, Dict

DEFAULT_INJECTOR_PUSHER_WAIT_TIME = 240
DEFAULT_MIN_PUSHER_WAIT_TIME = 5
DEFAULT_PUSHER_SECURITY_RATIO_TIME = 5
DEFAULT_PUSHER_MAX_RETRY = 3
DEFAULT_PUSHER_QUEUE_BATCH_SIZE = 100000  # we do not want too much broks on the socket in one pass

# Strings for log chapters
_MANAGE_BROKS_STR = get_chapter_string('MANAGE BROKS')
_GET_BROKS_STR = get_chapter_string('GET BROKS')
_RECEIVE_BROKS_STR = get_chapter_string('RECEIVE BROKS')
_BROKER_TIME_STR = get_chapter_string('BROKER TIME')
_MODULES_STR = get_chapter_string('MODULES')

CHAPTER_GET_BROKS = get_chapter_string('GET BROKS')
CHAPTER_RECEIVE_BROKS = get_chapter_string('RECEIVE BROKS')
CHAPTER_CONFIGURATION = get_chapter_string('CONFIGURATION')
CHAPTER_CONFIGURATION_UPDATE = get_chapter_string('CONFIGURATION-UPDATE')

_EXTERNAL_MODULE_STR = get_section_string('EXTERNAL MODULE')
_INTERNAL_MODULE_STR = get_section_string('INTERNAL MODULE')
_NEED_DATA_STR = get_section_string('NEED DATA')
_LOOP_START_STR = get_section_string('=== Loop start ===')
_LOOP_STOP_STR = get_section_string('=== Loop stop  ===')
_BUS_COMMANDS_STR = get_section_string('BUS COMMANDS')
_INTERNAL_COMMANDS_STR = get_section_string('INTERNAL COMMANDS')

_ARBITER_AS_BROKS_SOURCE = 'arbiter'


class IStatsBroker(IStatsInterface):
    """ 
    Interface for various stats about broker activity
    """
    
    doc = 'Return a connection status from my network location to all graphite server I should write to (graphite_perfdata type modules).'
    
    
    def check_graphite_write_status(self):
        return self.app.check_graphite_write_status()
    
    
    check_graphite_write_status.doc = doc
    check_graphite_write_status.need_lock = False
    
    doc = 'Return a connection status from my network location to all graphite server I should read from (webui type modules).'
    
    
    def check_graphite_read_status(self):
        return self.app.check_graphite_read_status()
    
    
    check_graphite_read_status.doc = doc
    check_graphite_read_status.need_lock = False
    
    
    def get_raw_stats(self, param=u''):
        # type: (unicode) -> Dict[unicode, Any]
        return super(IStatsBroker, self).get_raw_stats(param=param)
    
    
    get_raw_stats.doc = u'get stats of the daemon'
    get_raw_stats.need_lock = False
    
    
    def _daemon_get_raw_stats(self, param=u''):
        # type: (unicode) -> Dict[unicode, Any]
        app = self.app
        logger.info(u'Query broker running stats')
        raw_stats = {
            u'modules'              : [],
            u'len_external_commands': len(app.external_commands)
        }
        # update some queues
        with app.broks_lock:
            raw_stats[u'len_broks'] = len(app.broks)
        raw_stats[u'len_internal_broks'] = len(app.broks_internal_raised)
        with app.arbiter_broks_lock:
            raw_stats[u'len_arbiter_broks'] = len(app.arbiter_broks)
        
        instances = [inst for inst in app.modules_manager.get_all_instances() if inst.is_external]
        for inst in instances:
            try:
                raw_stats[u'modules'].append({u'module_name': inst.get_name(), u'queue_size': inst.to_q.qsize()})
            except Exception:  # noqa => need to catch all about queues
                raw_stats[u'modules'].append({u'module_name': inst.get_name(), u'queue_size': -1})
        
        raw_stats[u'module_stats'] = self._get_module_stats(getattr(self.app, u'modules_manager', None), param)
        raw_stats[u'http_errors_count'] = app.http_errors_count
        raw_stats[u'have_conf'] = app.cur_conf is not None
        raw_stats[u'activated'] = self.app.activated
        raw_stats[u'spare'] = self.app.spare
        raw_stats[u'master_daemon'] = self.app.master_daemon_name
        raw_stats[u'spare_daemon'] = self.app.spare_daemon_name
        raw_stats[u'spare_must_have_the_same_list_of_module_type'] = self.app.spare_must_have_the_same_list_of_module_type
        raw_stats[u'last_too_long_injector_time'] = self.app.last_too_long_injector_time
        raw_stats[u'known_realms'] = getattr(self.app, u'known_realms', None)
        return raw_stats


# Our main APP class
class Broker(WithInventorySatellite):
    properties = WithInventorySatellite.properties.copy()
    properties.update({
        'pidfile'                  : PathProp(default='brokerd.pid'),
        'port'                     : IntegerProp(default='7772'),
        'local_log'                : PathProp(default='brokerd.log'),
        'max_file_descriptor_limit': IntegerProp(default='0'),  # the broker need to set to maximum value available on the system
    })
    
    daemon_type = 'broker'
    
    
    def __init__(self, config_file, is_daemon, do_replace, debug, debug_file, profile='', daemon_id=0):
        super(Broker, self).__init__(self.__class__.daemon_type, config_file, is_daemon, do_replace, debug, debug_file, daemon_id)
        
        # Our arbiters
        self.arbiters = {}
        
        # Our pollers and reactionners
        self.pollers = {}
        self.reactionners = {}
        self.receivers = {}
        
        # Modules are load one time
        self.have_modules = False
        self.modules = None
        
        # Can have a queue of external_commands given by modules
        # will be processed by arbiter
        self.external_commands = []
        
        # All broks to manage
        self.broks = deque()  # broks to manage, note: WILL overwrite basesatellite value
        self.external_module_broks = deque()  # type:deque[Brok]  # broks manage by external module
        self.broks_lock = threading.RLock()  # as satellites will be access by threads, protect our broks list by a lock
        # broks raised this turn and that needs to be put in self.broks
        self.broks_internal_raised = []
        self.broks_packet_size = 0
        # broks raised by the arbiters, we need a lock so the push can be in parallel
        # to our current activities and won't lock the arbiter
        self.arbiter_broks = []
        self.arbiter_broks_lock = threading.RLock()
        
        self.timeout = 1.0
        
        self._add_http_interface(IStatsBroker(self))
        self._add_http_interface(IArbiterToInventorySatellite(self))
        
        self.local_module_stats = {}
        self.stats_time_sla = None
        self.stats_nb_sla = None
        self.broks_done = 0
        self.last_too_long_injector_time = 0.0
        
        # We will have a tread by distant satellites, so we must protect our access
        self.satellite_lock = threading.RLock()
        
        self._tmp_bucket = {}  # use by injector workers to inherit jobs during the fork()
        
        # Memory protection: we will have parameters to choose if we do memory protection or not
        self.manage_brok_enable_sub_processes_memory_usage_protection = False
        self.manage_brok_sub_process_memory_usage_system_reserved_memory = 0
        self.manage_brok_sub_processes_memory_usage_protection_max_retry_time = 5
        
        self.manage_brok_sub_process_broks_pusher_max_execution_timeout = DEFAULT_INJECTOR_PUSHER_WAIT_TIME
        self.manage_brok_sub_process_broks_pusher_security_ratio = DEFAULT_PUSHER_SECURITY_RATIO_TIME
        self.manage_brok_sub_process_broks_pusher_min_execution_timeout = DEFAULT_MIN_PUSHER_WAIT_TIME
        self.manage_brok_sub_process_broks_pusher_max_retry = DEFAULT_PUSHER_MAX_RETRY
        self.manage_brok_sub_process_broks_pusher_queue_batch_size = DEFAULT_PUSHER_QUEUE_BATCH_SIZE
        self.avg_brok_send_speed = AvgInRange(300)  # 0 broks/s by default, so we will take manage_brok_sub_process_broks_pusher_max_wait_time for wait time
        
        # Spare information
        self.spare_daemon_name = ''
        self.spare_must_have_the_same_list_of_module_type = True  # by default spare must have the same modules
        self.master_daemon_name = ''
        self.mainloop_watchdog = WatchDogThreadDumper(u'Main loop', wait_time=60 * 30, dump_interval=60 * 5, fatal_dead_lock_delay=60 * 30, multi_usage=True)
        
        # SEF-9050
        # register after fork cleanup for the broker daemon only yet
        # to generalize this processing to all daemons,
        # move next line to __init__ method of Daemon class
        after_fork_cleanup.register_top_instance(self)
    
    
    # BEWARE: you should have the lock to call this one
    def _find_scheduler_by_shard_id(self, shard_id):
        return next((scheduler for scheduler in self.schedulers.itervalues() if scheduler['instance_id'] == shard_id), None)
    
    
    # Schedulers have some queues. We can simplify the call by adding
    # elements into the proper queue just by looking at their type
    # Brok -> self.broks
    # External commands -> self.external_commands
    def add(self, elt):
        cls_type = elt.__class__.my_type
        if cls_type == 'brok':
            # Like this brok was set by scheduler 0
            elt.instance_id = 0
            self.broks_internal_raised.append(elt)
            return
        elif cls_type == 'externalcommand':
            logger.debug("[broker][add_broks] Enqueuing an external command '%s'" % str(ExternalCommand.__dict__))
            self.external_commands.append(elt)
        elif cls_type == 'message':
            # Maybe we got a Message from the modules, it's way to ask something like from now a full data from a scheduler for example.
            logger.debug('[broker][add_broks] message [%s]' % str(elt.__dict__))
            if elt.get_type() == 'NeedData':
                data = elt.get_data()
                
                if 'full_instance_id' not in data:
                    logger.warning("%s [ NEED DATA ] [%s] The asking for initial broks is malformed. (full_instance_id is missing)." % (CHAPTER_GET_BROKS, elt.source))
                    return
                
                shard_id = data['full_instance_id']
                source = elt.source
                with self.satellite_lock:
                    scheduler = self._find_scheduler_by_shard_id(shard_id)
                    # Maybe the arbiter did JUST remove the scheduler during the fact that the module get it, and now
                    # it's quite rare, but not an error because as soon as the scheduler will be given back by the arbiter
                    # we will have a full broks generation, so we don't have to "force" it
                    if scheduler is None:
                        logger.warning(
                            "[MODULES      ] [ NEED DATA ] [ %s ] The module is asking need to have all the elements from a scheduler. It is asking for shard (%s) but we cannot find it in our schedulers, because the arbiter did ask us to remove it. We will give it to the module as soon as the arbiter will give us this scheduler." % (
                                source, shard_id))
                        logger.warning("[MODULES      ] [ NEED DATA ] [ %s ] Current schedulers are: %s" % (source, ', '.join(['%s(shard=%s)' % (sched['name'], sched['instance_id']) for sched in self.schedulers.values()])))
                        return
                # Reset the scheduler, so it will ask for a full initial brok generation
                scheduler['con'] = None
                scheduler['daemon_incarnation'] = {}
                logger.info("%s [ NEED DATA ] [%s] Requesting scheduler '%s' (managing shard %s) initial broks." % (CHAPTER_GET_BROKS, source, scheduler['name'], shard_id))
            
            elif elt.get_type() == 'ICrash':
                # Maybe a module tells me that it's dead, I must log it's last words... and directly set the module to be restarted
                data = elt.get_data()
                module_name = data['name']
                instance = self.modules_manager.get_module_by_name(module_name)
                if instance is None:
                    logger.error('Did received an error from a unknown module: %s (%s)' % (module_name, data['trace']))
                    return
                module_crash_logger = LoggerFactory.get_logger('MODULES-MANAGER').get_sub_part('MODULE-INSTANCE-CRASH').get_sub_part(module_name).get_sub_part('module_type=%s' % instance.properties.get('type', ''))
                module_crash_logger.error('The module %s just stopped. Last ERROR received:' % module_name)
                if data['trace']:
                    for line in data['trace'].splitlines():
                        module_crash_logger.error('     %s' % line)
                # We can kill it
                self.modules_manager.did_crash(instance, reason=unicode(data['exception']), do_log=False)  # note: no more log here, already done ^^
    
    
    # Check if we do not connect to often to this
    def is_connection_try_too_close(self, elt):
        now = time.time()
        last_connection = elt['last_connection']
        if now - last_connection < 5:
            return True
        return False
    
    
    ##
    # Iterate over all graphite servers I should write to and return connexion status.
    # They are defined in my graphite_perfdata type modules.
    def check_graphite_write_status(self):
        reply = {u'status': u'OK', u'data': []}
        
        if not self.cur_conf:
            reply[u'status'] = u'NO_CONF'
            reply[u'data'].append({
                u'module_name'  : None,
                u'host'         : None,
                u'port'         : None,
                u'can_post_data': False
            })
            return reply
        
        graphite_mods = [mod for mod in self.cur_conf[u'global'][u'modules'] if mod.module_type == u'graphite_perfdata']
        if not graphite_mods:
            reply[u'status'] = u'NO_MODULE'
        for mod in graphite_mods:
            host = mod.host
            try:
                base_url = BaseUrl(host, mod.port or 2003)
                port = base_url.get_port()
                host = base_url.get_host()
            except ShinkenBaseUrlException:
                reply[u'status'] = u'CRITICAL'
                reply[u'data'].append({
                    u'module_name'  : mod.module_name,
                    u'host'         : mod.host,
                    u'port'         : mod.port,
                    u'can_post_data': False
                })
                return reply
            
            can_post_data = Broker.graphite_service_is_alive(host, port)
            
            reply[u'data'].append({
                u'module_name'  : mod.module_name,
                u'host'         : host,
                u'port'         : port,
                u'can_post_data': can_post_data})
        
        return reply
    
    
    ##
    # Iterate over all graphite servers I should read from and return connexion status.
    # They are defined in my webui type modules.
    def check_graphite_read_status(self):
        reply = []
        
        if not self.cur_conf:
            status = {
                u'module_name': None,
                u'host'       : None,
                u'ip'         : None,
                u'resolved_ip': None,
                u'port'       : None,
                u'use_ssl'    : None,
                u'realm'      : None,
                u'nb_metrics' : 0,
                u'reachable'  : False,
                u'cfg_url'    : u'',
                u'url_errors' : [],
                u'version'    : None
            }
            reply.append(status)
            return reply
        
        webui_modules = (mod for mod in self.cur_conf[u'global'][u'modules'] if mod.module_type == u'webui')
        webui_and_graphite_threads = []
        for webui_module in webui_modules:
            webui_graphite_backends = getattr(webui_module, u'graphite_backends', u'')
            graphite_backends = [g.strip() for g in webui_graphite_backends.split(u',') if g.strip()]
            use_ssl = to_bool(webui_module.use_ssl)
            graphs_errors = {}
            thread_name = 'webui-checks-errors-thread-%s' % webui_module.module_name
            webui_thread = threading.Thread(None, target=Broker.webui_graphs_errors, name=thread_name, args=(u'127.0.0.1', webui_module.port, use_ssl, graphs_errors))
            webui_and_graphite_threads.append(webui_thread)
            
            for graphite_backend in graphite_backends:
                graphite_backend_def = GraphiteBackend(graphite_backend, strict=False)
                if graphite_backend_def.has_errors():
                    status = {
                        u'module_name': webui_module.module_name,
                        u'host'       : None,
                        u'ip'         : None,
                        u'resolved_ip': None,
                        u'port'       : None,
                        u'use_ssl'    : None,
                        u'realm'      : graphite_backend_def.realm,
                        u'errors'     : None,
                        u'cfg_url'    : graphite_backend_def.cfg_url,
                        u'url_errors' : graphite_backend_def.get_errors(),
                        u'version'    : None
                    }
                    reply.append(status)
                    continue
                
                # This is run locally on broker, so this is the correct IP
                base_url = graphite_backend_def.base_url
                broker_addr = BaseUrl.get_local_ip()
                if base_url.is_localhost():
                    resolved_address = base_url.create_from(host=broker_addr)
                else:
                    resolved_address = base_url
                
                status = {
                    u'module_name': webui_module.module_name,
                    u'host'       : base_url.get_host(),
                    u'ip'         : base_url.get_ip(),
                    u'resolved_ip': resolved_address.get_ip(),
                    u'port'       : base_url.get_port(),
                    u'use_ssl'    : base_url.get_use_ssl(),
                    u'realm'      : graphite_backend_def.realm,
                    u'errors'     : graphs_errors,
                    u'cfg_url'    : graphite_backend_def.cfg_url,
                    u'url_errors' : [],
                    u'version'    : None
                }
                reply.append(status)
                thread_name = 'graphite-backend-thread-%s' % base_url.get_host()
                graphite_backend_thread = threading.Thread(None, target=Broker.graphite_ui_data, name=thread_name, args=(base_url, status))
                webui_and_graphite_threads.append(graphite_backend_thread)
        
        for t in webui_and_graphite_threads:
            t.daemon = True
            t.start()
        
        for t in webui_and_graphite_threads:
            t.join()
        
        return reply
    
    
    @staticmethod
    # Will return None in case of failure, or json reply.
    def graphite_ui_data(base_url, status, retry=4):
        # type: (BaseUrl, Dict, int) -> None
        Broker.get_graphite_version(base_url, status, retry)
        # We can only count nodes on our own versions, so check it's ok
        if not GraphiteVersionChecker().is_version_valid(status[u'version']):
            return
        while True:
            try:
                retry -= 1
                resp = urllib2.urlopen(base_url.build_url_with_path(u'/metrics/get-metrics-count'), timeout=3)
                buf = resp.read()
                json_data = json.loads(buf)
                status[u'nb_metrics'] = json_data.get(u'metrics', None)
                status[u'nb_hosts_clusters'] = json_data.get(u'level_0', None)
                status[GRAPHITE_STATS_KEY.TIME_READ] = json_data.get(GRAPHITE_STATS_KEY.TIME_READ, -1)
                status[GRAPHITE_STATS_KEY.LOCAL_TIME] = json_data.get(GRAPHITE_STATS_KEY.LOCAL_TIME, None)
                return
            except Exception:
                if retry == 0:
                    status[u'nb_metrics'] = 0
                    status[u'nb_hosts_clusters'] = 0
                    return
                else:
                    continue
    
    
    @staticmethod
    # Will return None in case of failure, or json reply.
    def get_graphite_version(base_url, status, retry=4):
        # type: (BaseUrl, Dict, int) -> None
        while True:
            try:
                retry -= 1
                resp = urllib2.urlopen(base_url.build_url_with_path(u'/version/'), timeout=3)
                buf = resp.read()
                status[u'reachable'] = True
                status[u'version'] = buf.strip()
                return
            except Exception:
                if retry == 0:
                    status[u'reachable'] = False
                    status[u'version'] = None
                    return
                else:
                    continue
    
    
    @staticmethod
    # Will return None in case of failure, or json reply.
    def webui_graphs_errors(host, port, use_ssl, graphs_errors, retry=4):
        while True:
            try:
                retry -= 1
                proto = u'https' if use_ssl else u'http'
                resp = urllib2.urlopen(u'%s://%s:%s/api/graphs/errors' % (proto, host, port), timeout=3)
                html = resp.read()
                jval = json.loads(html)
                graphs_errors.update(jval)
                return
            except Exception:
                if retry == 0:
                    return
                else:
                    continue
    
    
    @staticmethod
    def graphite_service_is_alive(host, port, retry=4):
        sock = None
        while True:
            try:
                retry -= 1
                sock = socket.create_connection((str(host), int(port)), 0.5)
                return True
            except Exception as exp:
                if retry == 0:
                    logger.warning("graphite_service_is_alive failed with exp: [%s]" % exp)
                    return False
                else:
                    continue
            finally:
                if sock is not None:
                    sock.close()
    
    
    # Get a brok. Our role is to put it in the modules
    # DO NOT CHANGE data of b!!!
    # REF: doc/broker-modules.png (4-5)
    def manage_brok(self, b):
        # Call all modules if they catch the call
        # data = b.data
        # if 'host' in b.type and 'group' not in b.type and "schedule" not in b.type:
        #         update = {"btype": b.type, "now": int(time.time()), "diff": int(time.time() - data.get('last_state_change',0))}
        #         update.update(data)
        #         try:
        #             logger.debug("[BROKSLE] %(now)s ; %(btype)20s ; %(host_name)10s ; %(last_state_id)s=>%(last_state_as_string)-10s ; %(state_id)s=>%(state)-10s ; %(last_state_change)s ; -%(diff)s\n" % update)
        #         except KeyError:
        #             logger.debug('[BROKSLE] ERROR for brok type %s' % b.type)
        
        self.broks_done += 1
        for mod in self.modules_manager.get_internal_instances():
            try:
                # Un-serialize if needed, but outside the module time (it's common to all modules)
                b.prepare()
                
                before = time.time()
                mod.manage_brok(b)
                mod_name = mod.get_name()
                if mod_name not in self.local_module_stats:
                    self.local_module_stats[mod_name] = 0
                self.local_module_stats[mod_name] += time.time() - before
            except Exception as exp:
                logger.debug(str(exp.__dict__))
                logger.warning("The mod %s raise an exception: %s, I'm tagging it to restart later" % (mod.get_name(), str(exp)))
                logger.warning("Exception type: %s" % type(exp))
                logger.warning("Back trace of this kill: %s" % (traceback.format_exc()))
                self.modules_manager.did_crash(mod, "The mod %s raise an exception: %s" % (mod.get_name(), str(exp)))
    
    
    # Add broks (a tab) to different queues for internal and external modules
    def _add_broks_to_queues(self, broks, source, duration=0.0, sent_broks=0, do_log=False):
        # type: (List[Brok], str, float, int, bool) -> None
        # If we are idle (spare waiting to start) we can't handle brok
        if not self.activated:
            return
        
        
        log_label = _RECEIVE_BROKS_STR if source == _ARBITER_AS_BROKS_SOURCE else _GET_BROKS_STR  # if we did get or not broks, we change the log
        
        # Ok now put in queue broks to be managed by
        # internal modules
        with self.broks_lock:
            self.broks.extend(broks)
            self.external_module_broks.extend(broks)
            nb_internal_broks = len(self.broks)  # Save the numbers in the with: part
            nb_external_broks = len(self.external_module_broks)
        
        # We only have to log now
        if not do_log:
            return
        
        logger.info('%s %s [PERF] [ %.3f ]s - Add %4s broks into INTERNAL queue (new size=%s) and the EXTERNAL queue (new size=%s)' % (log_label, source, duration, sent_broks, nb_internal_broks, nb_external_broks))

        # We don't show void list of types, if we have 0 broks, the fact that we really ask and ge nothing is enought (cf SEF-9763)
        if len(broks) != 0:
            number_by_types = {}
            for brok in broks:
                btype = brok.type
                number_by_types[btype] = number_by_types.get(btype, 0) + 1
            btypes = number_by_types.keys()
            btypes.sort()
            s_printed = ', '.join(['%s=%s' % (k, number_by_types[k]) for k in btypes])
            logger.info('%s %s                   ----- %4s composed of: %s' % (log_label, source, sent_broks, s_printed))
    
    
    # We will get in the broks list the broks from the arbiters,
    # but as the arbiter_broks list can be push by arbiter without Global lock,
    # we must protect this with the list lock
    def interger_arbiter_broks(self):
        with self.arbiter_broks_lock:
            self._add_broks_to_queues(self.arbiter_broks, _ARBITER_AS_BROKS_SOURCE)
            self.arbiter_broks = []
    
    
    # Get 'objects' from external modules
    # right now on nobody uses it, but it can be useful
    # for modules like livestatus to raise external
    # commands for example
    def get_objects_from_from_queues(self):
        start = time.time()
        nb_object_get = 0
        for module, queue in self.modules_manager.get_external_modules_and_from_queues():
            try:
                while not queue.empty():
                    o = queue.get(block=False)
                    self.add(o)
                    nb_object_get += 1
            except Exception:
                # NOTE: maybe the instance was crashing, and was stopped by a crash message in self.add(), it so, don't cry about it as
                # it is already die
                if self.modules_manager.is_instance_set_to_restart(module):
                    continue
                logger.warning('%s %s Cannot read shinken internal commands (like recheck, set acknowledge, etc) from module [%s]. We will retry it.' % (_MODULES_STR, _BUS_COMMANDS_STR, module.get_name()))
                logger.print_stack(level=logging.WARNING)
                continue
        end = time.time()
        logger.info('%s %s [PERF] [ %.3f ]s Did read %d shinken internal commands (like recheck, set acknowledge, etc) from modules' % (_MODULES_STR, _BUS_COMMANDS_STR, end - start, nb_object_get))
        return nb_object_get
    
    
    # For a new distant daemon, if it is a scheduler, ask for a new full broks generation
    def _manage_new_distant_daemon_incarnation(self, entry, old_incar, new_incar):
        daemon_type = entry['type']
        entry['broks'].clear()
        con = entry['con']
        if con is None:  # maybe another thread did close the connection on a new configuration
            entry['daemon_incarnation'] = {}  # reset it, so we will get back in this method on the next loop
            return
        # we must ask for a new full broks if  it's a scheduler
        if daemon_type == 'scheduler':
            if not new_incar:  # maybe the scheduler is currently void, wait until it have a real configuration loaded
                logger.info('%s %s [ %s ] Skipping initial broks generation of the scheduler that is currently void (old incarnation was %s)' % (_GET_BROKS_STR, _NEED_DATA_STR, entry['name'], old_incar))
                return
            logger.info('%s %s [ %s ] I ask for a initial broks generation to the scheduler with new daemon incarnation %s (old incarnation was %s)' % (_GET_BROKS_STR, _NEED_DATA_STR, entry['name'], new_incar, old_incar))
            
            self._add_broks_to_queues([Brok('asking_initial_broks', {'instance_id': entry['instance_id']})], 'internal broks')
            
            try:
                did_generate_raw = con.get('fill_initial_broks', {'bname': self.name}, wait='long')
            except HTTPExceptions as exp:  # If fail (as error 500 or something like this) it means that the
                # scheduler did fail to generate broks, it's a huge problem, and we must restart this
                msg = 'The scheduler %s did fail to generate initial broks data with an error: %s. we will retry it' % (entry['name'], exp)
                entry['daemon_incarnation'] = {}
                raise HTTPException(msg)
            # maybe the scheduler was not ready, if so, retry it
            # NOTE: we want all schedulers to be up-to-date, so ask for a real true here
            if did_generate_raw != 'true':
                msg = 'The scheduler %s did fail to generate initial broks data in this turn, retrying to ask it' % entry['name']
                entry['daemon_incarnation'] = {}
                raise HTTPException(msg)
    
    
    @staticmethod
    def __detect_protocol_and_get_input(input_data):
        # Protocol detection:
        # * 1 (old) was gzip
        # * 2 is a pickle with (protocol, seq_number,  data)
        try:
            input_data = SafeUnpickler.loads(input_data, u'broks')
            protocol = 2
        except cPickle.UnpicklingError:
            protocol = 1
        
        return protocol, input_data
    
    
    # We get new broks from schedulers
    # REF: doc/broker-modules.png (2)
    def get_new_broks(self, sat_entry):
        # If we are idle (spare waiting to start) we do not ask our satellites new broks
        if not self.activated:
            time.sleep(0.5)
            return
        # We check for new check in each scheduler and put the result in new_checks
        sat_type = sat_entry['type']
        try:
            con = sat_entry['con']  # type: HTTPClient
            if con is None:  # None = not initialized
                self.pynag_con_init(sat_entry)
                return  # we will get broks the next turn
            
            before = time.time()
            # Before ask a call that can be long, do a simple ping to be sure it is alive
            con.get('ping')
            scheduler_name = sat_entry['name']
            shard_id = u''
            configuration_incarnation_uuid = u''
            daemon_incarnation = sat_entry[u'daemon_incarnation']
            if daemon_incarnation:
                shard_id = daemon_incarnation.get(u'shard_id', u'')
                configuration_incarnation_uuid = daemon_incarnation.get(u'configuration_incarnation_uuid', u'')
            
            param = {
                u'bname'                         : self.name,
                u'packet_size'                   : self.broks_packet_size,
                u'seq_number'                    : sat_entry[u'sequence_number'],
                u'shard_id'                      : shard_id,
                u'configuration_incarnation_uuid': configuration_incarnation_uuid
            }
            tmp_broks_b64 = con.get(u'get_broks', param, wait=u'long')
            receive_time = time.time()
            input_data = base64.b64decode(tmp_broks_b64)
            del tmp_broks_b64  # release memory asap
            
            b64_time = time.time()
            
            protocol, input_data = self.__detect_protocol_and_get_input(input_data)
            nb_remaining_broks = 0
            # Now based on the protocol version, get broks as pickle
            try:
                if protocol == 2:
                    _protocol, seq_number, data, nb_remaining_broks = input_data
                    if _protocol != 2:
                        logger.error('%s [PROTOCOL] Invalid protocol version received from scheduler "%s" : %s' % (_GET_BROKS_STR, scheduler_name, _protocol))
                        sat_entry['con'] = None
                        return
                    pickle_broks = zlib.decompress(data)
                    sat_entry['sequence_number'] = seq_number
                else:
                    # protocol = 1
                    pickle_broks = zlib.decompress(input_data)
            except zlib.error as exp:
                logger.error('%s Fail to decompress loaded broks from %s %s with : [%s]' % (_GET_BROKS_STR, sat_type, scheduler_name, exp))
                sat_entry['con'] = None
                return
            
            uncompress_time = time.time()
            # Final Step: load the broks from pickle
            try:
                broks = SafeUnpickler.loads(pickle_broks, u'broks')
                del pickle_broks  # release memory asap
            except Exception as exp:
                logger.error('%s Fail to load broks data from %s %s with : [%s]' % (_GET_BROKS_STR, sat_type, sat_entry['name'], exp))
                sat_entry['con'] = None
                return
            unpickle_time = time.time()
            logger.debug('%s %s Got %4d broks, remaining broks: %4s broks total:%.3fs [recv:%.3fs, b64_decode:%.3fs, uncompress:%.3fs, unpickle:%.3fs]' % (
                _GET_BROKS_STR,
                sat_entry['name'],
                len(broks),
                nb_remaining_broks,
                unpickle_time - before,
                receive_time - before,
                b64_time - receive_time,
                uncompress_time - b64_time,
                unpickle_time - uncompress_time))
            
            sat_entry['sent_broks_this_turn'] += len(broks)
            
            # Ok, we can add theses broks to our queues
            now = time.time()
            # Log only if more than 1s since last log
            do_log = False
            if abs(now - sat_entry['last_log_time']) > 1:  # abs() is for time-jumps
                sat_entry['last_log_time'] = now
                do_log = True
            
            self._add_broks_to_queues(broks, sat_entry['name'], time.time() - before, sat_entry['sent_broks_this_turn'], do_log)
            sat_entry['sent_broks_this_turn'] = 0
            
            # We will re-ask broks if the scheduler say us that there is remaining broks, if not, wait normally
            loop_delay = 0 if nb_remaining_broks != 0 else 1
            sat_entry['loop_delay'] = loop_delay
        
        # Ok, con is not known, so we create it
        except KeyError as exp:
            logger.info("%s We fail to get new broks from %s %s with: [%s]. Connection must not be initialize. It will be initialize." % (_GET_BROKS_STR, sat_type, sat_entry['name'], str(exp)))
            self.pynag_con_init(sat_entry)
        except HTTPExceptions as exp:
            err_string = str(exp)
            did_log = False
            # IMPORTANT: to be removed ASAP we have broks transfer retries
            # NOTE: pycurl do not give us if the timeout was during the transfer or at the connect, so must look at the error part
            if 'Operation timed out after' in err_string:  # is a timeout
                if 'with 0 out of ' not in err_string:  # only raise a true error if there was a start of transfer
                    logger.error('*' * 100)
                    logger.error('%s Did fail to transfer broks from the %s "%s": [%s]. THESES BROKS ARE LOST AND CANNOT BE RETRIEVED.' % (_GET_BROKS_STR, sat_type, sat_entry['name'], str(exp)))
                    logger.error('*' * 100)
                    self._add_http_error_count()  # let the checks know that we have a HUGE errors here
                    did_log = True
            if not did_log:  # do not print a warning if it was a very ERROR
                logger.warning("%s We fail to get new broks from %s %s with: [%s]. Connection fail and it will be reinitialize." % (_GET_BROKS_STR, sat_type, sat_entry['name'], str(exp)))
            sat_entry['con'] = None
        # scheduler must not #be initialized
        except AttributeError as exp:
            logger.warning("%s We fail to get new broks from %s %s with: [%s]. The %s must be initialized." % (_GET_BROKS_STR, sat_type, sat_entry['name'], str(exp), sat_type))
        except Exception as exp:
            logger.error("%s We fail to get new broks from %s %s with: [%s]. Broker will be kill." % (_GET_BROKS_STR, sat_type, sat_entry['name'], str(exp)))
            logger.print_stack()
            sys.exit(1)
    
    
    # Helper function for module, will give our broks
    def get_retention_data(self):
        return {}
    
    
    # Get back our broks from a retention module
    def restore_retention_data(self, data):
        return
    
    
    def do_stop(self):
        self.print_log_block('Stopping daemon')
        self.daemon_info.daemon_is_requested_to_stop.value = True
        self.interrupted = True
        self.mainloop_watchdog.quit()
        super(Broker, self).do_stop()
    
    
    def _set_spare_daemon_name(self, spare_daemon_name, spare_must_have_the_same_list_of_module_type, _logger):
        # type: (str, bool, PartLogger) -> NoReturn
        previous_spare_daemon_name = self.spare_daemon_name
        self.spare_daemon_name = spare_daemon_name
        if previous_spare_daemon_name != self.spare_daemon_name:
            _logger.info('MASTER', 'My spare daemon is now "%s"' % self.spare_daemon_name)
        previous_spare_must_have_the_same_list_of_module_type = self.spare_must_have_the_same_list_of_module_type
        self.spare_must_have_the_same_list_of_module_type = spare_must_have_the_same_list_of_module_type
        if self.spare_daemon_name and previous_spare_must_have_the_same_list_of_module_type != self.spare_must_have_the_same_list_of_module_type:
            not_s = '' if self.spare_must_have_the_same_list_of_module_type else 'NOT '
            _logger.info('MASTER', 'The spare daemon "%s" is now %srequiring the same modules types as the master' % (self.spare_daemon_name, not_s))
    
    
    def _set_master_daemon_name(self, master_daemon_name, _logger):
        # type: (str, PartLogger) -> NoReturn
        previous_master_daemon_name = self.master_daemon_name
        self.master_daemon_name = master_daemon_name
        if previous_master_daemon_name != self.master_daemon_name:
            _logger.info('SPARE', 'I am now the spare of the master daemon "%s"' % self.master_daemon_name)
    
    
    # Set properties we want to set in our new schedulers
    def _set_default_values_to_scheduler_entry(self, entry):
        
        # IMPORTANT: mut be LOCAL, so each scheduler have their own {} and []
        default_scheduler_properties = {'broks'          : {}, 'thread': None, 'daemon_incarnation': {},
                                        'con'            : None, 'sequence_number': '', 'loop_delay': 1,
                                        'last_log_time'  : 0, 'sent_broks_this_turn': 0,
                                        'last_connection': 0, 'type': 'scheduler',
                                        }
        entry.update(default_scheduler_properties)
    
    
    def _set_daemon_id_of_scheduler(self, daemon, daemon_id):
        daemon['instance_id'] = daemon_id
    
    
    # setup a new conf, byt beware about global lock management.
    # Note: don't do locking thing here, as we have the satellite lock!
    def setup_new_conf(self):
        with self.satellite_lock:
            self.really_setup_new_conf()
    
    
    def really_setup_new_conf(self):
        _logger = self._print_new_update_conf_received()
        
        # If the configuration was giving us a new configuration incarnation, show it
        self.print_configuration_incarnation_log_entry_if_need(_logger)
        
        t0 = time.time()
        conf = self.new_conf
        self.new_conf = None
        self.cur_conf = conf
        # Got our name from the globals
        g_conf = conf['global']
        if 'broker_name' in g_conf:
            name = g_conf['broker_name']
        else:
            name = 'Unnamed broker'
        self.name = name
        logger.load_obj(self, name)
        
        self.broks_packet_size = g_conf.get('broks_packet_size', 2001024) * 1024
        self.manage_brok_enable_sub_processes_memory_usage_protection = g_conf.get('broker__manage_brok__enable_sub_processes_memory_usage_protection', True)
        self.manage_brok_sub_process_memory_usage_system_reserved_memory = g_conf.get('broker__manage_brok__sub_process_memory_usage_system_reserved_memory', 0)
        self.manage_brok_sub_processes_memory_usage_protection_max_retry_time = g_conf.get('broker__manage_brok__sub_processes_memory_usage_protection_max_retry_time', 5)
        
        self.manage_brok_sub_process_broks_pusher_max_execution_timeout = g_conf.get('broker__manage_brok__sub_process_broks_pusher_max_execution_timeout', DEFAULT_INJECTOR_PUSHER_WAIT_TIME)
        self.manage_brok_sub_process_broks_pusher_security_ratio = g_conf.get('broker__manage_brok__sub_process_broks_pusher_security_ratio', DEFAULT_PUSHER_SECURITY_RATIO_TIME)
        self.manage_brok_sub_process_broks_pusher_min_execution_timeout = g_conf.get('broker__manage_brok__sub_process_broks_pusher_min_execution_timeout', DEFAULT_MIN_PUSHER_WAIT_TIME)
        self.manage_brok_sub_process_broks_pusher_max_retry = g_conf.get('broker__manage_brok__sub_process_broks_pusher_max_retry', DEFAULT_PUSHER_MAX_RETRY)
        self.manage_brok_sub_process_broks_pusher_queue_batch_size = g_conf.get('broker__manage_brok__sub_process_broks_pusher_queue_batch_size', DEFAULT_PUSHER_QUEUE_BATCH_SIZE)
        
        # Look at human log format. Enable by default
        logger.set_human_format(on=g_conf.get('human_timestamp_log', True))
        
        self.save_daemon_name_into_configuration_file(name)
        
        # The arbiter let us know about the realms that are allowed to talk to us
        # it let us know also if a realm that was present before did disappear and so need to be deleted
        self.known_realms = conf['known_realms']
        
        spare_daemon_name = g_conf.get('spare_daemon', '')
        spare_must_have_the_same_list_of_module_type = g_conf.get('broker__manage_spare__spare_must_have_the_same_list_of_module_type', False)
        self._set_spare_daemon_name(spare_daemon_name, spare_must_have_the_same_list_of_module_type, _logger)
        self._set_master_daemon_name(g_conf.get('master_daemon', ''), _logger)
        
        self._set_spare(g_conf.get('spare', True), _logger)
        activated = conf.get('activated', True)
        was_activated = self._set_is_activated(activated, _logger)
        
        if not self.activated:
            self._go_as_not_active(was_activated, _logger)
            self.have_configuration = True
            return
        
        _logger.debug("Configuration received, set as active daemon")
        
        new_schedulers = []  # for logging
        deleted_schedulers = []
        new_arbiters = []
        
        # Load Schedulers
        for (daemon_id, daemon) in conf['schedulers'].iteritems():
            self._set_or_update_scheduler_from_configuration(daemon, daemon_id, g_conf, new_schedulers, deleted_schedulers, _logger)
        
        # Now get arbiter
        for arb_id in conf['arbiters']:
            # Must look if we already have it
            already_got = arb_id in self.arbiters
            if already_got:
                broks = self.arbiters[arb_id]['broks']
            else:
                broks = {}
            daemon = conf['arbiters'][arb_id]
            
            # replacing poller address and port by those defined in satellitemap
            # IMPORTANT: do this BEFORE get uri
            self._update_daemon_addr_port_with_satellitemap(daemon, g_conf['satellitemap'], _logger)
            uri = self._get_daemon_uri(daemon)
            daemon['uri'] = uri
            
            daemon['broks'] = broks
            daemon['instance_id'] = 0  # No use so all to 0
            daemon['daemon_incarnation'] = {}  # no use
            daemon['last_connection'] = 0
            daemon['id'] = arb_id
            daemon['type'] = 'arbiter'
            
            # We do not connect to the arbiter. Connection hangs
            if not already_got:
                new_arbiters.append((daemon['name'], daemon['uri']))
            self.arbiters[arb_id] = daemon
        
        self._print_new_and_deleted_daemons(new_schedulers=new_schedulers, deleted_schedulers=deleted_schedulers, new_arbiters=new_arbiters, _logger=_logger)
        
        if self.schedulers:
            self.modules = conf['global']['modules']
            [module.set_father_config(conf[u'global']) for module in self.modules]
            _logger.debug('Receiving modules:[%s] i already load modules:[%s]' % (','.join([m.get_name() for m in self.modules]), self.have_modules))
            
            if not self.have_modules:
                _logger.info('New modules received: %s' % (','.join([m.get_name() for m in self.modules])))
                # Ok now start, or restart them!
                # Set modules, init them and start external ones
                self.modules_manager.set_modules(self.modules)
                self.do_load_modules()
                self.modules_manager.start_external_instances()
                self.modules_manager.start_worker_based_instances()
                self.have_modules = True
            else:  # just update the one we need
                self.modules_manager.update_modules(self.modules)
            
            # Set our giving timezone from arbiter
            self.set_tz(conf['global']['use_timezone'])
            
            # Start threads if need, not a problem as starting thread is cheap and not timeout prone
            self.assert_valid_satellite_threads()
        
        self.have_configuration = True
        _logger.debug('configuration loaded into %.3fs' % (time.time() - t0))
    
    
    # An arbiter ask us to wait for a new conf, so we must clean
    # all our mess we did, and close modules too
    def clean_previous_run(self):
        with self.satellite_lock:
            # Clean all lists
            self.schedulers.clear()
            self.pollers.clear()
            self.reactionners.clear()
            self.receivers.clear()
        with self.broks_lock:
            self.broks = deque()
            self.external_module_broks = deque()
        self.broks_internal_raised = self.broks_internal_raised[:]
        with self.arbiter_broks_lock:
            self.arbiter_broks = self.arbiter_broks[:]
        self.external_commands = self.external_commands[:]
        
        # And now modules
        self.have_modules = False
        self.modules_manager.clear_instances()
    
    
    # Broker ask for Broks objects
    def get_jobs_from_distant(self, sat_entry):
        self.get_new_broks(sat_entry)
    
    
    def print_modules_stats(self, b_times, b_types, nb_broks):
        logger.debug('[broks] broks done this loop %d/%d' % (self.broks_done, nb_broks))
        stat_str = ''
        for mod in self.modules_manager.get_internal_instances():
            local_module_stats = self.local_module_stats.get(mod.get_name(), 0)
            if mod.get_name() == 'sla' and hasattr(mod, 'external_speed_counter'):
                
                if self.stats_time_sla is None:
                    self.stats_time_sla = AvgInRange(60, initial_value=mod.external_speed_counter)
                    self.stats_nb_sla = AvgInRange(60, initial_value=mod.external_done_counter)
                else:
                    self.stats_time_sla.update_avg(mod.external_speed_counter)
                    self.stats_nb_sla.update_avg(mod.external_done_counter)
                
                avg_speed = self.stats_time_sla.get_avg(mod.external_speed_counter)
                avg = self.stats_nb_sla.get_avg(mod.external_done_counter)
                stat_str += '[%s][%.4fs update_time:%.4f, nb_done:%4d, speed:%4d /avg update_time:%.4f, nb_done:%4d, speed:%4d ]  |  ' % \
                            ('sla',
                             local_module_stats,
                             mod.external_speed_counter,
                             mod.external_done_counter,
                             0 if mod.external_speed_counter == 0 else mod.external_done_counter / mod.external_speed_counter,
                             avg_speed,
                             avg,
                             0 if avg_speed == 0 else avg / avg_speed)
                mod.external_speed_counter = 0
                mod.external_done_counter = 0
            elif mod.get_name() == 'Graphite-Perfdata' and hasattr(mod, 'external_info'):
                stat_str += '[%s][%.4fs %s]  |  ' % (mod.get_name(), local_module_stats, mod.external_info)
            else:
                stat_str += '[%s][%.4fs]  |  ' % (mod.get_name(), local_module_stats)
            if statsd is not None:
                statsd.timing('broker.%s' % mod.get_name(), local_module_stats)
        logger.debug('[broks] stats by local module %s' % stat_str)
        stat_str = ''
        for brok_type, brok_stat_nb in b_types.iteritems():
            brok_stat_time = b_times[brok_type]
            stat_str += '{ [%s] : [%d][%.4f]s }' % (brok_type, brok_stat_nb, brok_stat_time)
        logger.debug('[broks] stats by type %s' % stat_str)
    
    
    def print_external_modules_queue_size(self):
        external_modules = self.modules_manager.get_external_instances()
        if not external_modules:
            return
        msg = [u'%s %s => Number of "Broks Sets" not eaten in MODULE queues' % (_MANAGE_BROKS_STR, _EXTERNAL_MODULE_STR)]
        for external_module in external_modules:
            try:
                msg.append(u' (%s): %s  ' % (external_module.get_name(), external_module.to_q.qsize()))
            except Exception as exp:
                msg.append(u' (%s): Exception! %s   ' % (external_module.get_name(), exp))
        logger.info(u''.join(msg))
    
    
    @staticmethod
    def _in_injector_send_to_one_queue(queue, broks):
        logger.debug(u'Starting to send %d broks to the module' % len(broks))
        try:
            t0 = time.time()
            queue.put(broks, timeout=DEFAULT_INJECTOR_PUSHER_WAIT_TIME)  # we put a timeout so we won't lock forever
            logger.debug(u'[ PERF ] %d broks were sent in %.3fs' % (len(broks), time.time() - t0))
        except Exception as exp:
            # As we did have a raw fail, we need to have the serialization size for debug
            # NOTE: it takes a bit of time, but should be ok with the timeout system
            broks_total_size_ko = int(float(len(cPickle.dumps(broks, cPickle.HIGHEST_PROTOCOL))) / 1024)
            logger.error(u'The %s broks (total size=%s Ko) sent did fail with the error : %s' % (len(broks), broks_total_size_ko, exp))
            # reraise the error, so we have a traceback and a forced exit
            raise
    
    
    @staticmethod
    def _in_injector_look_at_father(fatherpid):
        import os
        import time
        while True:
            try:
                os.kill(fatherpid, 0)  # fake kill
                time.sleep(0.5)
            except:  # noqa => no more father? fuck!
                os._exit(0)  # noqa => in a thread, raw kill
    
    
    def _in_queue_injector(self, before_start, module_name, father_pid):
        try:
            real_start = time.time()
            # We will need to
            # PERFORMANCE HACK/CHOICE:
            # IMPORTANT: we need to do as the multiprocessing lib is doing manually as
            #            we did fork() and not call the Process that is too slow for us
            
            # BEWARE: if we directly call the multiprocessing.util._run_after_forkers()
            #         it will be too long (500ms!) for EACH fork. it's not possible.
            #         it seems that only the multiprocessing.managers.AutoProxy[Queue]._after_for
            #         are long (500ms each) because they are doing a connection, so we will only
            #         clean the only queue we are using.
            # BEWARE2: do not try to thread clean phase calls, I did try, and it dead lock :(
            (injector_queue, broks) = self._tmp_bucket[module_name]
            self._tmp_bucket.clear()  # free the memory as soon as possible
            auto_proxy_queue_class = type(injector_queue)  # get the queue class, because it can move/rename in the future
            
            # Code take from multiprocessing.Process._bootstrap
            multiprocessing.util._finalizer_registry.clear()  # noqa => yes we know about this hack
            #  => code from : multiprocessing.util._run_after_forkers()
            items = list(multiprocessing.util._afterfork_registry.items())  # noqa => yes we know about this hack
            items.sort()
            for (index, ident, func), obj in items:
                # only clean the queue we will use in this process
                if isinstance(obj, auto_proxy_queue_class) and obj != injector_queue:
                    continue
                try:
                    func(obj)
                except Exception:
                    pass
            
            injector_start_time = time.time()
            process_name = u"%s [ - Module: %s - queue brok pusher ]" % (self.daemon_display_name, module_name)
            set_process_name(process_name)
            logger_name = process_name + u'(pid=%s)' % os.getpid()
            logger_name = logger_name.replace(u'[', u'(').replace(u']', u')')
            logger.set_name(logger_name, log_this_change=False)  # change the logger, but don't put a INFO line about it
            logger.debug(u'The worker process was started in %.3fs (clean phase=%.3f)' % (time.time() - before_start, injector_start_time - real_start))
            
            # NOTE: cannot call the clean_previous_run because it needs locks, and we are not sure threads lock
            # are not release in this fork().
            self.broks = deque()
            self.external_module_broks = deque()
            
            # Classic sub process behavior/protection
            # start_malloc_trim_thread()  # NOT this one, it's too long for a short time worker (~200ms)
            look_at_my_father_thread = LookAtMyFatherThread(father_pid, self.daemon_display_name, process_name, loop_speed=0.5)
            look_at_my_father_thread.start_thread()
            
            # Ask the http to release the socket
            self.http_daemon.shutdown(quiet=True)
            
            # Do not want the daemon anymore from here
            self.new_conf = None
            self.cur_conf = None
            if hasattr(self, 'conf'):
                del self.conf
            
            # Allow the memory dump
            self.set_exit_handler()
            
            logger.debug(u'Finish to initialize the injector worker in %.3fs' % (time.time() - injector_start_time))
            # Ok, now work..
            self._in_injector_send_to_one_queue(injector_queue, broks)
            # Important: the broker NEED to know that all was OK, so give a return code = 0
            os._exit(0)  # noqa => raw death
        except Exception:  # noqa => raw death
            logger.error(u'Cannot send broks to the instance %s: %s' % (module_name, traceback.format_exc()))
            os._exit(2)  # noqa => raw death
            # IMPORTANT: the broker NEED to know that the push did fail, so give a return code != 0
    
    
    def _get_avg_speed(self):
        return self.avg_brok_send_speed.get_avg(flatten=False, default_value=0)
    
    
    # We will wait for the pusher process based on the broks send speed
    # * avg speed * size of broks to send = expected_time
    # Then we will multiply to allow more security and not kill process when we can avoid
    # * expected_time * security = secure_expected_time
    # Maybe we have very few broks, but the fork() have a minimal time, so at least will be a security value
    # * min(DEFAULT_MIN_PUSHER_WAIT_TIME, secure_expected_time) = ceiled_secure_expected_time
    def _get_pusher_wait_time(self, nb_broks_sent):
        avg_speed = self._get_avg_speed()
        if avg_speed == 0:
            logger.info('%s %s We do not have any broks spend speed, so will allow a wait of %.1fs for sending %d broks' % (_MANAGE_BROKS_STR, _EXTERNAL_MODULE_STR, self.manage_brok_sub_process_broks_pusher_max_execution_timeout, nb_broks_sent))
            return self.manage_brok_sub_process_broks_pusher_max_execution_timeout
        expected_time = nb_broks_sent / avg_speed
        secure_expected_time = expected_time * self.manage_brok_sub_process_broks_pusher_security_ratio
        ceiled_secure_expected_time = min(self.manage_brok_sub_process_broks_pusher_max_execution_timeout, max(secure_expected_time, self.manage_brok_sub_process_broks_pusher_min_execution_timeout))
        
        logger.debug('%s %s We will allow a wait of %.1fs for sending %d broks (average speed = %5d broks/s)' % (_MANAGE_BROKS_STR, _EXTERNAL_MODULE_STR, ceiled_secure_expected_time, nb_broks_sent, avg_speed))
        return ceiled_secure_expected_time
    
    
    def _wait_for_all_injectors_to_finish(self, all_processes, elapsed_put_stats, nb_broks_sent, nb_try):
        not_finish = [e for e in all_processes.values() if e['end'] is None]
        wait_start = time.time()
        
        errored_modules = []  # we give push failed that will be retry by caller if max try not reached
        
        ceiled_secure_expected_time = self._get_pusher_wait_time(nb_broks_sent)
        
        warning_printed = False
        
        nb_try_str = ' [ TRY: %s / %s ]' % (nb_try, self.manage_brok_sub_process_broks_pusher_max_retry)
        
        while len(not_finish) != 0:
            for e in all_processes.values():
                if e['end'] is not None:  # already finished
                    continue
                module = e['module']
                module_name = module.get_name()
                pid = e['pid']
                last_warning = e['last_warning']  # do not raise warnings too much
                now = time.time()
                # raise a warning about long process, but at least once a second
                if now > wait_start + (ceiled_secure_expected_time / 2) and now > last_warning + 1:
                    if not warning_printed:
                        logger.warning((u'%s %s [PERF-ALERT] [%s] |' % (_MANAGE_BROKS_STR, _EXTERNAL_MODULE_STR, module_name)) + (u'-' * 98) + u'|')
                        warning_printed = True
                    logger.warning(u'%s %s [PERF-ALERT] [%s] The queue pushing process (pid=%s) is running since a very long time: %ds' % (_MANAGE_BROKS_STR, _EXTERNAL_MODULE_STR, module_name, pid, now - wait_start))
                    e['last_warning'] = now
                    self.last_too_long_injector_time = now  # let the daemon stats know about this error
                
                # If VERY too long, just kill it
                if now > wait_start + ceiled_secure_expected_time:
                    self.last_too_long_injector_time = now  # let the daemon stats know about this error
                    logger.warning(u'%s %s [PERF-ALERT] [%s] The queue pushing process (pid=%s) seems to be in a dead lock state (running since %ds>%ds limit). Stopping the process.' % (
                        _MANAGE_BROKS_STR, _EXTERNAL_MODULE_STR, module_name, pid, now - wait_start, ceiled_secure_expected_time))
                    # kill the injector process
                    try:
                        os.kill(pid, signal.SIGKILL)
                        os.waitpid(pid, 0)
                    except OSError:
                        pass
                    e['end'] = now  # do not reenter in this kill part
                    errored_modules.append(module_name)  # so the caller will retry it if possible
                    continue
                
                # we need to know what the exit status of the pusher, and as we did fork() we will have a system raw
                # return code, that we need to transform into a classic return code (like 2)
                killed_pid, rc = -1, 0
                try:
                    os.kill(pid, 0)
                    killed_pid, rc = os.waitpid(pid, os.WNOHANG)
                    os.kill(pid, 0)
                except OSError:  # Not alive
                    # just to be sure it's dead-dead
                    try:
                        os.kill(pid, signal.SIGKILL)
                        os.waitpid(pid, 0)
                    except OSError:
                        pass
                    
                    e['end'] = time.time()  # in all cases we did finish with this one
                    if rc == 0:  # if the pusher did exit normally, just save the stats
                        elapsed_put = e['end'] - e['start']
                        broks_speed = nb_broks_sent / elapsed_put
                        logger.debug(u'%s %s [%s] The queue injector (pid=%s) did finished in %.3fs (speed=%s broks/s)' % (_MANAGE_BROKS_STR, _EXTERNAL_MODULE_STR, module_name, pid, elapsed_put, broks_speed))
                        self.avg_brok_send_speed.update_avg(broks_speed)
                        elapsed_put_stats[module_name] = elapsed_put_stats.get(module_name, 0) + elapsed_put
                    else:  # oups, pusher did have a hard fail, we must set this module in error mode, like a timeout case
                        try:  # transform raw eit code into more classic one
                            display_rc = os.WEXITSTATUS(rc)
                        except:  # can be in error if the process is traced or something strange like this
                            display_rc = u'(unknown, raw rc=%s)' % rc
                        logger.warning(
                            u'%s %s %s [%s] The queue pushing process (pid=%s) did fail to send Broks with en a error return code %s. Retrying if possible.' % (_MANAGE_BROKS_STR, _EXTERNAL_MODULE_STR, nb_try_str, module_name, pid, display_rc))
                        errored_modules.append(module_name)  # so the caller will retry it if possible
                        continue
            
            not_finish = [e for e in all_processes.values() if e['end'] is None]
            if len(not_finish) != 0:
                time.sleep(0.01)
        
        return errored_modules
    
    
    def _push_broks_to_external_queues(self, broks, elapsed_put_stats):
        # We are sending broks as a big list, more efficient than one by one
        modules_and_queues = self.modules_manager.get_external_modules_and_queues()
        
        all_processes = {}
        self._tmp_bucket.clear()
        
        queues_jobs = {}
        for (instance, queue) in modules_and_queues:
            queues_jobs[instance.get_name()] = (instance, queue)
        
        try_nb = 0
        while len(queues_jobs) != 0:
            try_nb += 1
            if try_nb > self.manage_brok_sub_process_broks_pusher_max_retry:
                for module_name in queues_jobs.keys():
                    ceiled_secure_expected_time = self._get_pusher_wait_time(len(broks))
                    err = u'%s %s [PERF-ALERT] [%s] The module did failed to send broks after %d try with a %.1fs execution timeout -- min=%d(seconds) < %d broks / %d (speed in broks/s) * %d (security ratio) < max=%ds .' % (
                        _MANAGE_BROKS_STR, _EXTERNAL_MODULE_STR, module_name, self.manage_brok_sub_process_broks_pusher_max_retry, ceiled_secure_expected_time,
                        self.manage_brok_sub_process_broks_pusher_min_execution_timeout, len(broks), self._get_avg_speed(),
                        self.manage_brok_sub_process_broks_pusher_security_ratio, self.manage_brok_sub_process_broks_pusher_max_execution_timeout,
                    )
                    logger.error(err)
                    logger.error((u'%s %s [PERF-ALERT] [%s] |' % (_MANAGE_BROKS_STR, _EXTERNAL_MODULE_STR, module_name)) + ('-' * 30) + ' Killing and restarting the module ' + ('-' * 30) + '|')
                    self.modules_manager.did_crash(queues_jobs[module_name][0], reason=err)  # instance
                    # won't be possible to send this one, remove it
                    del queues_jobs[module_name]
                break
            
            for (instance, queue) in modules_and_queues:
                # before_put = time.time()
                module_name = instance.get_name()
                self._tmp_bucket[module_name] = (queue, broks)
                if self.manage_brok_enable_sub_processes_memory_usage_protection:
                    fork_is_possible = mem_wait_for_fork_possible(u'brok pusher for %s' % module_name,
                                                                  retry_time=self.manage_brok_sub_processes_memory_usage_protection_max_retry_time,
                                                                  reserved_memory=self.manage_brok_sub_process_memory_usage_system_reserved_memory)
                    if not fork_is_possible:
                        err = u'[PERFORMANCE] [MEMORY] There are not enough memory to start the brok pusher process for module %s. Killing the module and restarting it.' % module_name
                        self.modules_manager.did_crash(instance, reason=err)
                        # won't be possible to send this one, remove it
                        del queues_jobs[module_name]
                        continue
                before = time.time()
                fatherpid = os.getpid()
                try:
                    pid = os.fork()
                except OSError as exp:  # Maybe the fork can fail, like for memory issue if changed since tested
                    err = u'[MANAGE BROKS ] [ EXTERNAL MODULE    ] Cannot launch the brok pusher for module %s ("%s"). Killing the module and restarting it.' % (module_name, exp)
                    self.modules_manager.did_crash(instance, reason=err)
                    # won't be possible to send this one, remove it
                    del queues_jobs[module_name]
                    continue
                
                if pid == 0:  # the son
                    self._in_queue_injector(before, module_name, fatherpid)  # will exit
                # Father only:
                entry = {'module': instance, 'pid': pid, 'start': time.time(), 'end': None, 'last_warning': 0.0}
                all_processes[instance.get_name()] = entry
            
            # We will wait until all injectors are done
            not_finished_modules = self._wait_for_all_injectors_to_finish(all_processes, elapsed_put_stats, len(broks), try_nb)
            
            # if some errored modules occur, we don't remove them
            finished_modules = [module_name for module_name in queues_jobs.keys() if module_name not in not_finished_modules]
            for module_name in finished_modules:
                # SEF-4969: The module might have been removed from the queue by another thread in the meantime
                queues_jobs.pop(module_name, None)
        
        # We no more need the tmp bucket now, all was sent
        self._tmp_bucket.clear()
    
    
    def do_loop_turn(self):
        # If the loop is NOT finished after 30min, it's a problem, and a FATAL one. Dump stack every 5minutes
        # until the daemon is restarted
        # NOTE: 30min because in the past we did have long bug with 20min time, so take a delay to only have
        #       real dead locks
        with self.mainloop_watchdog:
            self._do_loop_turn()
    
    
    def _do_loop_turn(self):
        start_snap = cpu_stats_helper.get_thread_cpu_snapshot()
        loop_start = time.time()
        self._increase_loop_number()
        loop_number = self._get_loop_number()
        logger.info(_BROKER_TIME_STR)
        logger.info('%s %s [ Loop number=%-5d ] ===-===-===-===-===-===-===-===-===-===-===-===-===' % (_BROKER_TIME_STR, _LOOP_START_STR, loop_number))
        
        # If there are some zombie process, wait for all of them and allow them to go to hell
        try:
            active_children()
        except:
            pass
        
        # #### STEP1:
        
        with self.broks_lock:
            nb_broks = len(self.broks)
        logger.debug("[broks][manage] Begin loop with [%d] broks to handles" % nb_broks)
        
        self.broks_done = 0
        self.local_module_stats.clear()  # be sure to remove old unused modules
        for mod in self.modules_manager.get_internal_instances():
            self.local_module_stats[mod.get_name()] = 0
        
        # Dump modules Queues size
        self.print_external_modules_queue_size()
        
        # Begin to clean modules
        self.check_and_del_zombie_modules()
        
        # Look if we have all need threads for our satellites
        self.assert_valid_satellite_threads()
        
        # Maybe the arbiter ask us to wait for a new conf
        # If true, we must restart all...
        if self.cur_conf is None:
            # Clean previous run from useless objects and close modules
            self.clean_previous_run()
            
            self.wait_for_initial_conf()
            # we may have been interrupted or so; then just return from this loop turn
            if not self.new_conf:
                return
            
            self.setup_new_conf()
        
        # Now we check if arbiter speak to us in the pyro_daemon.
        # If so, we listen for it
        # When it pushes conf to us, we re-initialise connections
        self.watch_for_new_conf(0.0)
        if self.new_conf:
            self.setup_new_conf()
        
        # We need to be sure that our inventory do not keep deleted realms
        self._clean_old_realms_in_inventory()
        
        # If an inventory did change, warn the modules about it so they can update their own inventory about it
        self.assert_module_inventory_are_updated()
        
        if int(time.time()) % 3600 == 0:
            for (realm_name, inventory) in self._realms_inventory.items():
                logger.info('The realm %s inventory have currently %s elements' % (realm_name, inventory.get_len()))
        
        # Reap broks sent from the arbiters
        self.interger_arbiter_broks()
        
        # Spare mode stop loop here
        if not self.activated:
            # we clear broks list as in idle (spare waiting to start) we do not handle broks
            with self.broks_lock:
                self.broks = deque()
                self.external_module_broks = deque()
            
            logger.info('[BROKER TIME  ] Broker is idle (spare waiting to start)')
            logger.info('[BROKER TIME  ] [ === Loop stop === ] [ Loop number=%-5d ] ===-===-===-===-===-===-===-===-===-===-===-===-===' % loop_number)
            logger.debug('[BROKER TIME  ] [ === Loop stop === ] [ Loop number=%-5d ] %s' % (loop_number, start_snap.get_diff()))
            
            loop_time = (time.time() - loop_start)
            self.sleep(1.0 - loop_time)
            
            return
        
        # #### STEP2: PREPARE and send broks to external queues
        
        step2_start = time.time()
        # EXTERNAL BROKS
        # We will works this turn with a copy of the broks, so we won't be impacted by the others threads
        with self.broks_lock:
            broks = copy.copy(self.broks)  # type: deque[Brok]
            broks_to_send_to_externals = list([b for b in self.external_module_broks if b.type != 'log'])
            
            self.broks = deque()
            self.external_module_broks = deque()
        logger.info('%s %s [PERF] [ %.3f ]s, preparing broks lists for INTERNAL and EXTERNAL modules' % (_MANAGE_BROKS_STR, get_section_string('PREPARING BROKS'), time.time() - step2_start))
        
        # and for external queues
        # REF: doc/broker-modules.png (3)
        # We put to external queues broks that was not already send
        before_external = time.time()
        # We are sending broks as a big list, more efficient than one by one
        modules_and_queues = self.modules_manager.get_external_modules_and_queues()
        
        elapsed_put_stats = {}
        nb_external_broks_at_start = len(broks_to_send_to_externals)
        nb_external_broks_managed = 0
        
        while len(broks_to_send_to_externals) != 0:
            now = time.time()
            if now < before_external or now > before_external + 1:  # if get back in time, or too long, skip this turn
                # so we must re-merge our last broks with the main broks to do not lost them
                with self.broks_lock:
                    # IMPORTANT: extendleft does REVERSE broks orders so we do once before to put them in good order
                    # note: broks.reverse() is not available on python 2.6
                    self.external_module_broks.extendleft(reversed(broks_to_send_to_externals))
                break
            
            # Take the next batch of max self.manage_brok_sub_process_broks_pusher_queue_batch_size
            shard_to_send = broks_to_send_to_externals[:self.manage_brok_sub_process_broks_pusher_queue_batch_size]
            broks_to_send_to_externals = broks_to_send_to_externals[self.manage_brok_sub_process_broks_pusher_queue_batch_size:]
            self._push_broks_to_external_queues(shard_to_send, elapsed_put_stats)
            nb_external_broks_managed += len(shard_to_send)
        
        # logger.info('[MANAGE BROKS ] [ EXTERNAL MODULE    ] We will allow a wait of %.1fs for sending %d broks (average speed = %5d broks/s)' % (ceiled_secure_expected_time, nb_broks_sent, avg_speed))
        ceiled_secure_expected_time = self._get_pusher_wait_time(nb_external_broks_at_start)
        avg_speed = self._get_avg_speed()
        logger.info('%s %s - PUSHED   [ %.3fs, limit=%.3fs ]s, EXTERNAL queue evolution: [ %5d broks => %5d broks remaining ] [ %5d broks managed ] [ Push average speed = %5d broks/s]' % (
            _MANAGE_BROKS_STR, _EXTERNAL_MODULE_STR, time.time() - before_external, ceiled_secure_expected_time, nb_external_broks_at_start, len(self.external_module_broks), nb_external_broks_managed, avg_speed))
        if len(elapsed_put_stats) != 0:
            elapsed_puts_sorted = sorted(elapsed_put_stats.keys())
            logger.info('%s                            ----- Sent to [%d] EXTERNAL queues: %s' % (_MANAGE_BROKS_STR, len(modules_and_queues), ', '.join(['(%s=%.3fs)' % (k, elapsed_put_stats[k]) for k in elapsed_puts_sorted])))
        
        # #### STEP3: MANAGE internal broks
        start_step3 = time.time()
        
        # Now INTERNAL
        b_types = defaultdict(int)
        b_times = defaultdict(float)
        
        nb_broks_managed = 0
        nb_broks_at_start = len(broks)
        
        brok_prepare_time = 0
        
        internal_modules = self.modules_manager.get_internal_instances()
        have_internal_modules = len(internal_modules) != 0
        if have_internal_modules:
            while len(broks) != 0:
                now = time.time()
                
                # Do not 'manage' more than 1s, we must get new broks
                # every 1s
                if now - start_step3 > 1:
                    # so we must re-merge our last broks with the main broks to do not
                    # lost them
                    with self.broks_lock:
                        # IMPORTANT: extendleft does REVERSE broks orders so we do once before to put them in good order
                        # note: broks.reverse() is not available on python 2.6
                        self.broks.extendleft(reversed(broks))
                    break
                
                try:
                    b = broks.popleft()
                except IndexError:  # no more broks, maybe a daemon stop, not a problem, catch it
                    break
                
                nb_broks_managed += 1
                before_prepare = time.time()
                b.prepare()
                brok_prepare_time = brok_prepare_time + (time.time() - before_prepare)
                # Ok, we can get the brok, and doing something with it
                # REF: doc/broker-modules.png (4-5)
                # We un serialize the brok before consume it
                before_manage = time.time()
                self.manage_brok(b)
                after_manage = time.time()
                b_types[b.type] += 1
                b_times[b.type] += (after_manage - before_manage)
        else:  # no internal modules, we can fake the result
            nb_broks_managed = len(broks)
        
        with self.broks_lock:
            nb_remaining_broks = len(self.broks)
        internal_broks_time = time.time() - start_step3
        if have_internal_modules:
            logger.info('%s %s - EXECUTED [ %.3f ]s, INTERNAL queue evolution: [ %5d broks => %5d broks remaining ] [ %5d broks managed ]' % (
                _MANAGE_BROKS_STR, _INTERNAL_MODULE_STR, internal_broks_time, nb_broks_at_start, nb_remaining_broks, nb_broks_managed))
        
        sorted_modules_stats = self.local_module_stats.keys()
        sorted_modules_stats.sort()
        if sorted_modules_stats:
            logger.info('%s                            ----- Details of INTERNAL brok deserialization time:%.3fs modules execution time: %s' % (
                _MANAGE_BROKS_STR,
                brok_prepare_time,
                ', '.join(['(%s=%.3fs)' % (k, self.local_module_stats[k]) for k in sorted_modules_stats])))
        
        # #### STEP4: get modules objects and call modules ticks
        # Maybe external modules raised 'objects' we should get them
        self.get_objects_from_from_queues()
        
        # Say to modules it's a new tick :)
        self.hook_point('tick')
        
        self.print_modules_stats(b_times, b_types, nb_broks)
        
        loop_time = (time.time() - loop_start)
        
        logger.info('%s %s [ Loop number=%-5d ] [PERF] [ %.3f ]s' % (_BROKER_TIME_STR, _LOOP_STOP_STR, loop_number, loop_time))
        logger.debug('%s %s [ Loop number=%-5d ] [PERF] %s' % (_BROKER_TIME_STR, _LOOP_STOP_STR, loop_number, start_snap.get_diff()))
        self.sleep(1.0 - loop_time)
    
    
    #  Main function, will loop forever
    def main(self):
        try:
            self.load_config_file()
            
            for line in self.get_header():
                logger.info(line)
            self.daily_log_version()
            
            logger.info("[broks] Using working directory: %s" % os.path.abspath(self.workdir))
            
            # Look if we are enabled or not. If ok, start the daemon mode
            self.look_for_early_exit()
            self.do_daemon_init_and_start()
            self.print_daemon_start()
            self.load_modules_manager()
            
            self._register_http_interfaces()
            
            #  We wait for initial conf
            self.wait_for_initial_conf()
            if not self.new_conf:
                return
            
            self.setup_new_conf()
            
            # Now the main loop
            self.do_mainloop()
        except Exception:
            logger.critical("The daemon did have an unrecoverable error. It must exit.")
            logger.critical("You can log a bug to your Shinken integrator with the error message:")
            logger.critical("%s" % (traceback.format_exc()))
            os._exit(2)  # noqa => force process exit, skip all hooks
            raise
    
    
    def do_after_fork_cleanup(self, after_fork_new_top_instance):
        # SEF-9050
        # Running after fork cleanup
        # We are no more in this daemon process,
        # its son (this process) need to free useless resources from father
        super(Broker, self).do_after_fork_cleanup(after_fork_new_top_instance)
        self.arbiters = {}
        self.pollers = {}
        self.reactionners = {}
        self.receivers = {}
        self.schedulers = {}
        self.modules = None
        self.external_commands = []
        self.broks = None
        self.external_module_broks = None
        self.broks_internal_raised = []
        self.arbiter_broks = []
        self.local_module_stats = {}
        self.stats_time_sla = None
        self.stats_nb_sla = None
        self._tmp_bucket = {}
        self.new_conf = None
        self.cur_conf = None
